#include <c64.h>
#include <stdbool.h>
#include <string.h>
#include <peekpoke.h>
#include <conio.h>
#include "gfx.h"
#include "edit.h"


#define SWAP(a, b) do { (a) ^= (b); (b) ^= (a); (a) ^= (b); } while ( 0 )
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define ABSDIFF(a, b) ((a) < (b) ? ((b) - (a)) : ((a) - (b)))


static unsigned char border_color;


static void init_hires(void) {
    border_color = VIC.bordercolor;
    VIC.bordercolor = 0;
    CIA2.pra = 2;       // gfx bank $4000
    VIC.ctrl1 = 0x3b;   // bitmap mode
    VIC.ctrl2 = 0xc8;   // hires
    VIC.addr = 0x90;    // bitmap at $4000, screen at $6400
    // fill altscreen colors with white on black
    memset(altscreen, 0x10, sizeof(altscreen));
    
    // initialize cursor sprites
    memcpy(&bitmap[0x1f40], curspr, sizeof(curspr));
    POKE(&screen[1016], 0x7d);
    POKE(&altscreen[1016], 0x7d);
    POKE(&screen[1017], 0x7e);
    POKE(&altscreen[1017], 0x7e);
    POKE(&screen[1018], 0x7f);
    POKE(&altscreen[1018], 0x7f);
    VIC.spr_ena = 1;
    VIC.spr_exp_y = 0;
    VIC.spr_bg_prio = 2;
    VIC.spr_mcolor = 0;
    VIC.spr_exp_x = 0;
}

static void exit_hires(void) {
    VIC.bordercolor = border_color;
    CIA2.pra = 3;       // gfx bank $0000
    VIC.ctrl1 = 0x1b;   // text mode
    VIC.addr = 0x17;    // lower case rom font
    VIC.spr_ena = 0;    // disable sprites
}


static void flip(signed char x, signed char y, unsigned char w, unsigned char h) {
    unsigned int i, line;
    unsigned char *bmpptr;
    unsigned char *scrptr;
    
    bmpptr = &bitmap[y * 320 + x * 8];
    scrptr = &screen[y * 40 + x];
    for (line = 0; line < h; ++line) {
        for (i = 0; i < w; ++i) {
            scrptr[i] = (scrptr[i] << 4) | (scrptr[i] >> 4);
        }
        for (i = 0; i != w * 8; ++i) {
            bmpptr[i] = bmpptr[i] ^ 0xff;
        }
        bmpptr += 320;
        scrptr += 40;
    }
}


unsigned char sprxpos(unsigned char x) {
    return (x * 8 + 23) & 0xff;
}

unsigned char sprypos(unsigned char y) {
    return y * 8 + 49;
}


void edit(void) {
    unsigned char c;
    unsigned char fill = false;
    signed char x = 0, y = 0, fx = 0, fy = 0;
    
    init_hires();
    
    for (;;) {
        VIC.spr0_x = VIC.spr1_x = sprxpos(x);
        VIC.spr0_y = VIC.spr1_y = sprypos(y);
        if (x >= 30) {
            VIC.spr_hi_x |= 3;
        } else {
            VIC.spr_hi_x &= 0xfc;
        }
        
        while (!kbhit()) {
            VIC.spr2_color = VIC.spr1_color = ++VIC.spr0_color;
        }
        c = cgetc();
        switch (c) {
            
            // run/stop to exit editor
            case 0x03:
            exit_hires();
            return;
            break;
            
            // <- to toggle bitmap display
            case 0x5f:
            VIC.addr = VIC.addr ^ 0x10;
            VIC.spr_ena = VIC.spr_ena ^ 2;
            break;
            
            // x to set corner for fill
            case 'x':
            fx = x;
            fy = y;
            fill = true;
            VIC.spr2_x = sprxpos(fx);
            VIC.spr2_y = sprypos(fy);
            if (fx > 30) {
                VIC.spr_hi_x |= 4;
            } else {
                VIC.spr_hi_x &= 0xfb;
            }
            VIC.spr_ena |= 4;
            break;
            
            // space to flip low and hi nybble
            case 0x20:
            if (fill) {
                flip(MIN(x, fx), MIN(y, fy), ABSDIFF(x, fx) + 1, ABSDIFF(y, fy) + 1);
                fill = false;
                VIC.spr_ena &= 0xfb;
            } else {
                flip(x, y, 1, 1);
            }
            break;
            
            // cursor keys to move cursor
            // down
            case 0x11:
            if (++y > 24) {
                y = 0;
            }
            break;
            // right
            case 0x1d:
            if (++x > 39) {
                x = 0;
                if (++y > 24) {
                    y = 0;
                }
            }
            break;
            // up
            case 0x91:
            if (--y < 0) {
                y = 24;
            }
            break;
            // left
            case 0x9d:
            if (--x < 0) {
                x = 39;
                if (--y < 0) {
                    y = 24;
                }
            }
            break;
            
        }
    }
}
